import inspect
import weakref
from collections import Mapping, Sequence
from functools import reduce
import contextlib
from typing import (
    Any, Callable, Dict, Iterable, TypeVar, Union, get_type_hints
)

from ._utils import get_arguments_specification
from .container import DependencyContainer
from .injector import DependencyInjector
from .providers import FactoryProvider, ParameterProvider

T = TypeVar('T')


class DependencyManager:
    """Provides utility functions/decorators to manage dependencies.

    Except for :py:meth:`attrib()` all functions can either be used as
    decorators or functions to directly modify an object.

    Custom instances or classes can be used as :py:attr:`container` and
    :py:attr:`injector`.

    """

    auto_wire = True  # type: bool
    """
    Default value for :code:`auto_wire` argument in methods such as
    :py:meth:`register()` or :py:meth:`factory()`
    """

    use_names = False  # type: bool
    """
    Default value for :code:`use_names` argument in methods such as
    :py:meth:`inject()` or :py:meth:`register()`
    """

    arg_map = None  # type: Dict
    """
    Default mapping for :code:`arg_map` argument in methods such as
    :py:meth:`inject()` or :py:meth:`register()`.
    """

    def __init__(self,
                 auto_wire: bool = None,
                 use_names: bool = None,
                 arg_map: Mapping = None,
                 container: DependencyContainer = None,
                 injector: DependencyInjector = None
                 ) -> None:
        """Initialize the DependencyManager.

        Args:
            auto_wire: Default value for :code:`auto_wire` argument.
            use_names: Default value for :code:`use_names` argument.
            container: Container to use if specified.
            injector: Injector to use if specified.

        """
        if auto_wire is not None:
            self.auto_wire = auto_wire
        if use_names is not None:
            self.use_names = use_names
        self.arg_map = dict()
        self.arg_map.update(arg_map or dict())

        self.container = container or DependencyContainer()
        self.container[DependencyContainer] = weakref.proxy(self.container)

        self.injector = injector or DependencyInjector(self.container)
        self.container[DependencyInjector] = weakref.proxy(self.injector)

        self.provider(FactoryProvider, auto_wire=False)
        self._factories = (
            self.container.providers[FactoryProvider]
        )  # type: FactoryProvider
        self.provider(ParameterProvider, auto_wire=False)
        self._parameters = (
            self.container.providers[ParameterProvider]
        )  # type: ParameterProvider

    def __repr__(self):
        return (
            "{}(auto_wire={!r}, mapping={!r}, use_names={!r}, "
            "container={!r}, injector={!r})"
        ).format(
            type(self).__name__,
            self.auto_wire,
            self.arg_map,
            self.use_names,
            self.container,
            self.injector
        )

    @property
    def providers(self):
        return self.container.providers

    def inject(self,
               func: Callable = None,
               arg_map: Union[Mapping, Sequence] = None,
               use_names: Union[bool, Iterable[str]] = None,
               bind: bool = False
               ) -> Callable:
        """Inject dependencies into the function.

        Args:
            func: Callable for which the argument should be injected.
            arg_map: Custom mapping of the arguments name to their respective
                dependency id. A sequence of dependencies can also be
                specified, which will be mapped to the arguments through their
                order. Annotations are overriden.
            use_names: Whether the arguments name should be used to find for
                a dependency. An iterable of names may also be provided to
                restrict this to a subset of the arguments. Annotations are
                overriden, but not the arg_map.
            bind: bind arguments with :py:func:`functools.partial` directly to
                avoid the overhead of retrieving dependencies. This should only
                be used when necessary, it makes testing a lot more difficult.

        Returns:
            The injected function or a decorator.

        """
        if use_names is None:
            use_names = self.use_names

        def _inject(f):
            _mapping = self.arg_map.copy()
            if isinstance(arg_map, Mapping):
                _mapping.update(arg_map)
            elif isinstance(arg_map, Sequence):
                arg_spec, _, _ = get_arguments_specification(f)
                for (name, _), dependency_id in zip(arg_spec, arg_map):
                    _mapping[name] = dependency_id

            if bind:
                return self.injector.bind(func=f, arg_map=_mapping,
                                          use_names=use_names)

            return self.injector.inject(func=f, arg_map=_mapping,
                                        use_names=use_names)

        return func and _inject(func) or _inject

    def register(self,
                 cls: type = None,
                 singleton: bool = True,
                 auto_wire: Union[bool, Iterable[str]] = None,
                 arg_map: Union[Mapping, Sequence] = None,
                 use_names: Union[bool, Iterable[str]] = None
                 ) -> Union[Callable, type]:
        """Register a dependency by its class.

        Args:
            cls: Class to register as a dependency. It will be instantiated
                only when requested.
            singleton: If True, the class will be instantiated only once,
                further will receive the same instance.
            auto_wire: Injects automatically the dependencies of the methods
                specified, or only of :code:`__init__()` if True.
            arg_map: Custom mapping of the arguments name to their respective
                dependency id. A sequence of dependencies can also be
                specified, which will be mapped to the arguments through their
                order. Annotations are overriden.
            use_names: Whether the arguments name should be used to find for
                a dependency. An iterable of names may also be provided to
                restrict this to a subset of the arguments. Annotations are
                overriden, but not the arg_map.

        Returns:
            The class or the class decorator.

        """
        if auto_wire is None:
            auto_wire = self.auto_wire

        def register_class(_cls):
            if not inspect.isclass(_cls):
                raise ValueError(
                    "Expecting a class, found a {}".format(type(_cls))
                )

            if auto_wire:
                _cls = self.wire(
                    _cls,
                    methods=(
                        ('__init__',) if auto_wire is True else auto_wire
                    ),
                    arg_map=arg_map,
                    use_names=use_names
                )

            self._factories.register(dependency_id=_cls, factory=_cls,
                                     singleton=singleton)

            return _cls

        return cls and register_class(cls) or register_class

    def factory(self,
                func: Callable = None,
                dependency_id: Any = None,
                auto_wire: Union[bool, Iterable[str]] = None,
                singleton: bool = True,
                arg_map: Union[Mapping, Sequence] = None,
                use_names: Union[bool, Iterable[str]] = None,
                build_subclasses: bool = False
                ) -> Callable:
        """Register a dependency providers, a factory to build the dependency.

        Args:
            func: Callable which builds the dependency.
            dependency_id: Id of the dependency. Defaults to the return type of
                :code:`func` if specified.
            singleton: If True the dependency_provider is called only once.
                Otherwise it is called anew every time.
            auto_wire: If :code:`func` is a function, its dependencies are
                injected if True. Should :code:`func` be a class with
                :py:func:`__call__`, dependencies of :code:`__init__()` and
                :code:`__call__()` will be injected if True. One may also
                provide an iterable of method names requiring dependency
                injection.
            arg_map: Custom mapping of the arguments name to their respective
                dependency id. A sequence of dependencies can also be
                specified, which will be mapped to the arguments through their
                order. Annotations are overriden.
            use_names: Whether the arguments name should be used to find for
                a dependency. An iterable of names may also be provided to
                restrict this to a subset of the arguments. Annotations are
                overriden, but not the arg_map.
            build_subclasses: If True, subclasses will also be build with this
                factory. If multiple factories are defined, the first in the
                MRO is used.

        Returns:
            object: The dependency_provider

        """
        if auto_wire is None:
            auto_wire = self.auto_wire

        def register_factory(factory):
            if inspect.isclass(factory):
                if '__call__' not in dir(factory):
                    raise ValueError("Factory class needs to be callable.")
                type_hints = get_type_hints(factory.__call__) or {}

                if auto_wire:
                    factory = self.wire(
                        factory,
                        methods=(
                            ('__call__', '__init__')
                            if auto_wire is True else
                            auto_wire
                        ),
                        arg_map=arg_map,
                        use_names=use_names
                    )

                factory = factory()
            else:
                if not callable(factory):
                    raise ValueError("factory parameter needs to be callable.")
                type_hints = get_type_hints(factory) or {}

                if auto_wire:
                    factory = self.inject(factory,
                                          arg_map=arg_map,
                                          use_names=use_names)

            self._factories.register(
                factory=factory,
                singleton=singleton,
                dependency_id=dependency_id or type_hints.get('return'),
                build_subclasses=build_subclasses
            )
            return factory

        return func and register_factory(func) or register_factory

    def wire(self,
             cls: type = None,
             methods: Iterable[str] = None,
             **inject_kwargs
             ) -> Union[Callable, type]:
        """Wire a class by injecting the dependencies in all specified methods.

        Args:
            cls: class to wire.
            methods: Name of the methods for which dependencies should be
                injected. Defaults to all defined methods.
            **inject_kwargs: Keyword arguments passed on to :py:meth:`inject`.

        Returns:
            type: Wired class.

        """

        def wire_methods(_cls):
            nonlocal methods

            if methods is None:
                methods = map(
                    lambda m: m[0],  # get only the name
                    inspect.getmembers(
                        _cls,
                        # Retrieve static methods, class methods, methods.
                        predicate=lambda f: (inspect.isfunction(f)
                                             or inspect.ismethod(f))
                    )
                )

            for method in methods:
                setattr(_cls,
                        method,
                        self.inject(getattr(_cls, method), **inject_kwargs))

            return _cls

        return cls and wire_methods(cls) or wire_methods

    def attrib(self,
               dependency_id: Any = None,
               use_name: bool = None,
               **attr_kwargs
               ):
        """Injects a dependency with attributes defined with attrs package.

        Args:
            dependency_id: Id of the dependency to inject. Defaults to the
                annotation.
            use_name: If True, use the attribute name as the dependency id
                overriding any annotations.
            **attr_kwargs: Keyword arguments passed on to attr.ib()

        Returns:
            object: attr.Attribute with a attr.Factory.

        """
        try:
            import attr
        except ImportError:
            raise RuntimeError("attrs package must be installed.")

        if use_name is None:
            use_name = self.use_names

        def factory(instance):
            nonlocal dependency_id

            if dependency_id is None:
                cls = instance.__class__
                type_hints = get_type_hints(cls) or {}

                for attribute in attr.fields(cls):
                    # Dirty way to find the attrib annotation.
                    # Maybe attr will eventually provide the annotation ?
                    if isinstance(attribute.default, attr.Factory) \
                            and attribute.default.factory is factory:
                        try:
                            dependency_id = type_hints[attribute.name]
                        except KeyError:
                            if use_name:
                                dependency_id = attribute.name
                                break
                        else:
                            break
                else:
                    raise ValueError(
                        "No dependency could be detected. Please specify "
                        "the parameter `dependency_id` or `use_name=True`."
                        "Annotations may also be used."
                    )

            return self.container[dependency_id]

        return attr.ib(default=attr.Factory(factory, takes_self=True),
                       **attr_kwargs)

    def provider(self,
                 cls: type = None,
                 auto_wire: Union[bool, Iterable[str]] = None,
                 arg_map: Union[Mapping, Sequence] = None,
                 use_names: Union[bool, Iterable[str]] = None
                 ) -> Union[Callable, type]:
        """Register a providers by its class.

        Args:
            cls: class to register as a provider. The class must have a
                :code:`__antidote_provide()` method accepting as first argument
                the dependency id. Variable keyword and positional arguments
                must be accepted as they may be provided.
            auto_wire: If True, the dependencies of :code:`__init__()` are
                injected. An iterable of method names which require dependency
                injection may also be specified.
            arg_map: Custom mapping of the arguments name to their respective
                dependency id. A sequence of dependencies can also be
                specified, which will be mapped to the arguments through their
                order. Annotations are overriden.
            use_names: Whether the arguments name should be used to find for
                a dependency. An iterable of names may also be provided to
                restrict this to a subset of the arguments. Annotations are
                overriden, but not the arg_map.

        Returns:
            the providers's class or the class decorator.
        """
        if auto_wire is None:
            auto_wire = self.auto_wire

        def register_provider(_cls):
            if not inspect.isclass(_cls):
                raise ValueError("Expecting a class, "
                                 "found {}".format(repr(type(_cls))))

            if not hasattr(_cls, '__antidote_provide__'):
                raise ValueError("Method __antidote_provide__() "
                                 "must be defined")

            if auto_wire:
                _cls = self.wire(
                    _cls,
                    methods=(
                        ('__init__',) if auto_wire is True else auto_wire
                    ),
                    arg_map=arg_map,
                    use_names=use_names
                )

            self.container.providers[_cls] = _cls()

            return _cls

        return cls and register_provider(cls) or register_provider

    def register_parameters(self,
                            parameters: T,
                            getter: Union[Callable[[T, Any], Any], str] = None,
                            prefix: str = '',
                            split: str = ''
                            ) -> Callable:
        """
        Register a mapping of parameters and its associated parser.

        Args:
            parameters: Object containing the parameters, usually a mapping.
            getter: Function retrieving the dependency from the parameters, it
                must accept the parameters as first argument and the dependency
                id as second.
            prefix: If specified, only string prefixed with it will be taken
                into account.
            split: If specified, only string dependency_id are accepted, which
                will be split by :code:`split`. The result is used to
                recursively retrieve the dependency from :code:`parameters`.
                Beware that :py:exc:`TypeError` will be converted to
                :py:exc:`LookUpError` and thus be ignored, as the recursion may
                go too far.

        Returns:
            getter callable or decorator.
        """
        if not isinstance(prefix, str) or not isinstance(split, str):
            raise ValueError("prefix and split arguments must be strings.")

        def register_parser(f):
            if not callable(f):
                raise ValueError("parser must be callable or be 'getitem'")

            if prefix or split:
                def _getter(params, dependency_id):
                    if not isinstance(dependency_id, str) \
                            or not dependency_id.startswith(prefix):
                        raise LookupError(dependency_id)

                    dependency_id = dependency_id[len(prefix):]

                    if split:
                        try:
                            return reduce(f, dependency_id.split(split),
                                          params)
                        except TypeError as e:
                            raise LookupError(dependency_id) from e

                    return f(params, dependency_id)
            else:
                _getter = f

            self._parameters.register(parameters, _getter)

            return f

        return getter and register_parser(getter) or register_parser

    @contextlib.contextmanager
    def context(self,
                dependencies: Union[Mapping, Iterable] = None,
                include: Iterable = None,
                exclude: Iterable = None,
                missing: Iterable = None
                ):
        """
        Creates a context within one can control which of the defined
        dependencies available or not. Any changes will be discarded at the
        end.

        >>> from antidote import antidote, DependencyContainer
        >>> with antidote.context(include=[]):
        ...     s = 'Your code isolated from every other dependencies'
        ...     antidote.container[DependencyContainer]
        <... DependencyContainer ...>

        The :py:class:`~antidote.DependencyInjector` and the
        :py:class:`~antidote.DependencyContainer` will still be accessible.

        Args:
            dependencies: Dependencies instances used to override existing ones
                in the new context.
            include: Iterable of dependency to include. If None
                everything is accessible.
            exclude: Iterable of dependency to exclude.
            missing: Iterable of dependency which should raise a
                :py:exc:`~.exceptions.DependencyNotFoundError` even if a
                provider could instantiate them.

        """
        with self.container.context(dependencies=dependencies, include=include,
                                    exclude=exclude, missing=missing):
            # Re-inject DependencyManager's globals
            self.container[DependencyContainer] = weakref.proxy(self.container)
            self.container[DependencyInjector] = weakref.proxy(self.injector)
            yield



from .__version__ import __version__
from .container import Instance, DependencyContainer, Dependency
from .exceptions import (
    DependencyCycleError, DependencyDuplicateError, AntidoteError,
    DependencyInstantiationError, DependencyNotFoundError,
    DependencyNotProvidableError
)
from .injector import DependencyInjector
from .manager import DependencyManager

__all__ = [
    'Instance',
    'DependencyContainer',
    'DependencyInjector',
    'DependencyManager',
    'AntidoteError',
    'DependencyNotProvidableError',
    'DependencyNotFoundError',
    'DependencyDuplicateError',
    'DependencyCycleError',
    'DependencyInstantiationError',
    'Dependency',
    'antidote'
]

antidote = DependencyManager()



from typing import Any, Callable, TypeVar

from ..container import Dependency, Instance
from ..exceptions import DependencyNotProvidableError

T = TypeVar('T')


class ParameterProvider:
    """
    Provider managing constant parameters like configuration.
    """

    def __init__(self):
        self._parameter_getter_couples = []

    def __repr__(self):
        return "{}(parameter_getter_couples={!r})".format(
            type(self).__name__,
            tuple(param for param, _ in self._parameter_getter_couples)
        )

    def __antidote_provide__(self, dependency: Dependency) -> Instance:
        """
        Provide the parameter associated with the dependency_id.

        Args:
            dependency: dependency to provide.

        Returns:
            A :py:class:`~.container.Instance` wrapping the built instance for
            the dependency.
        """
        for parameters, getter in self._parameter_getter_couples:
            try:
                parameter = getter(parameters, dependency.id)
            except LookupError:
                pass
            else:
                return Instance(parameter, singleton=True)

        raise DependencyNotProvidableError(dependency)

    def register(self, parameters: T, getter: Callable[[T, Any], Any]):
        """
        Register parameters with its getter.

        Args:
            getter: Function used to retrieve a requested dependency from the
                parameters. It must have similar signature to
                :py:func:`~operator.getitem` accepting as first argument
                the object from which to retrieve the data and as second the
                key.
            parameters: Object containing the parameters, usually a mapping.

        """
        self._parameter_getter_couples.append((parameters, getter))



class AntidoteError(Exception):
    """ Base class of all errors of antidote. """


class DependencyDuplicateError(AntidoteError):
    """
    A dependency already exists with the same id.
    *May* be raised by providers.
    """


class DependencyInstantiationError(AntidoteError):
    """
    The dependency could not be instantiated.
    Raised by the container.
    """


class DependencyCycleError(AntidoteError):
    """
    A dependency cycle is found.
    Raised by the container.
    """


class DependencyNotProvidableError(AntidoteError):
    """
    The dependency could not be provided.
    Raised by providers.
    """


class DependencyNotFoundError(AntidoteError):
    """
    The dependency could not be found in the container.
    Raised by the container.
    """



import pytest

from antidote import Instance, DependencyContainer, Dependency
from antidote import (
    DependencyCycleError, DependencyInstantiationError,
    DependencyNotFoundError, DependencyNotProvidableError
)


class DummyProvider(object):
    singleton = True

    def __init__(self, data=None):
        self.data = data or dict()

    def __setitem__(self, key, value):
        self.data[key] = value

    def __antidote_provide__(self, dependency_id, *args, **kwargs):
        try:
            return Instance(self.data[dependency_id],
                            singleton=self.singleton)
        except KeyError:
            raise DependencyNotProvidableError(dependency_id)


class DummyFactoryProvider(object):
    create_singleton = True

    def __init__(self, data=None):
        self.data = data or dict()

    def __setitem__(self, key, value):
        self.data[key] = value

    def __antidote_provide__(self, dependency_id, *args, **kwargs):
        try:
            return Instance(self.data[dependency_id](),
                            singleton=self.create_singleton)
        except KeyError:
            raise DependencyNotProvidableError(dependency_id)


class Service(object):
    def __init__(self, *args):
        pass


class AnotherService(object):
    def __init__(self, *args):
        pass


class YetAnotherService(object):
    def __init__(self, *args):
        pass


class ServiceWithNonMetDependency(object):
    def __init__(self, dependency):
        pass


@pytest.fixture()
def container():
    return DependencyContainer()


def test_dependency_repr():
    o = object()
    d = Instance(o, False)

    assert repr(False) in repr(d)
    assert repr(o) in repr(d)


def test_setitem(container: DependencyContainer):
    s = object()
    container['service'] = s

    assert s is container['service']
    assert repr(s) in repr(container)


def test_update(container: DependencyContainer):
    container[Service] = Service()

    another_service = AnotherService()
    x = object()

    container.update({
        Service: another_service,
        'x': x
    })

    assert another_service is container[Service]
    assert x is container['x']


def test_extend(container: DependencyContainer):
    provider = DummyProvider()
    container.providers[DummyProvider] = provider

    # Can't check directly if the container is the same, as a proxy is used
    assert provider is container.providers[DummyProvider]


def test_getitem_and_provide(container: DependencyContainer):
    container.providers[DummyFactoryProvider] = DummyFactoryProvider({
        Service: lambda: Service(),
        ServiceWithNonMetDependency: lambda: ServiceWithNonMetDependency(),
    })
    container.providers[DummyProvider] = DummyProvider({'name': 'Antidote'})

    for get in (container.__getitem__, container.provide):
        with pytest.raises(DependencyNotFoundError):
            get(object)

        with pytest.raises(DependencyInstantiationError):
            get(ServiceWithNonMetDependency)

    assert isinstance(container[Service], Service)
    assert isinstance(container[Dependency(Service)], Service)
    assert isinstance(container.provide(Service), Service)
    assert 'Antidote' == container['name']
    assert 'Antidote' == container[Dependency('name')]
    assert 'Antidote' == container.provide('name')


def test_singleton(container: DependencyContainer):
    container.providers[DummyFactoryProvider] = DummyFactoryProvider({
        Service: lambda: Service(),
        AnotherService: lambda: AnotherService(),
    })

    service = container[Service]
    container.providers[DummyFactoryProvider][Service] = lambda: object()
    assert service is container[Service]

    container.providers[DummyFactoryProvider].create_singleton = False
    another_service = container[AnotherService]
    assert another_service is not container[AnotherService]


def test_cycle_error(container: DependencyContainer):
    container.providers[DummyFactoryProvider] = DummyFactoryProvider({
        Service: lambda: Service(container[AnotherService]),
        AnotherService: lambda: AnotherService(container[YetAnotherService]),
        YetAnotherService: lambda: YetAnotherService(container[Service]),
    })

    with pytest.raises(DependencyCycleError):
        container[Service]

    with pytest.raises(DependencyCycleError):
        container[AnotherService]

    with pytest.raises(DependencyCycleError):
        container[YetAnotherService]


def test_repr_str(container: DependencyContainer):
    container.providers[DummyProvider] = DummyProvider({'name': 'Antidote'})
    container['test'] = 1

    assert 'test' in repr(container)
    assert repr(container.providers[DummyProvider]) in repr(container)
    assert str(container.providers[DummyProvider]) in str(container)


def test_context_isolation(container: DependencyContainer):
    container['test'] = 1
    container['name'] = 'Antidote'
    s = Service()
    container[Service] = s

    with container.context():
        assert s is container[Service]
        assert 1 == container['test']
        assert 'Antidote' == container['name']

        container['another_service'] = AnotherService()
        assert isinstance(container['another_service'], AnotherService)

        s2 = Service()
        container[Service] = s2
        assert s2 is container[Service]

    with pytest.raises(DependencyNotFoundError):
        container['another_service']

    assert s is container[Service]
    assert 1 == container['test']
    assert 'Antidote' == container['name']


def test_context_include(container: DependencyContainer):
    container['test'] = 1
    container['name'] = 'Antidote'
    s = Service()
    container[Service] = s

    with container.context(include=[Service]):
        assert s is container[Service]

        with pytest.raises(DependencyNotFoundError):
            container['name']

        with pytest.raises(DependencyNotFoundError):
            container['test']

    with container.context(include=[]):
        with pytest.raises(DependencyNotFoundError):
            container[Service]

        with pytest.raises(DependencyNotFoundError):
            container['name']

        with pytest.raises(DependencyNotFoundError):
            container['test']


def test_context_exclude(container: DependencyContainer):
    container['test'] = 1
    container['name'] = 'Antidote'
    s = Service()
    container[Service] = s

    with container.context(exclude=['name']):
        assert s is container[Service]
        assert 1 == container['test']

        with pytest.raises(DependencyNotFoundError):
            container['name']


def test_context_override(container: DependencyContainer):
    container['test'] = 1
    container['name'] = 'Antidote'
    s = Service()
    container[Service] = s

    with container.context(dependencies=dict(test=2, name='testing')):
        assert s is container[Service]
        assert 2 == container['test']
        assert 'testing' == container['name']


def test_context_missing(container: DependencyContainer):
    container['test'] = 1
    container.providers[DummyProvider] = DummyProvider({'name': 'Antidote'})
    s = Service()
    container[Service] = s

    with container.context(missing=['name']):
        assert s is container[Service]
        assert 1 == container['test']

        with pytest.raises(DependencyNotFoundError):
            container['name']

    with container.context(missing=['test'], include=[Service]):
        assert s is container[Service]
        assert 'Antidote' == container['name']

        with pytest.raises(DependencyNotFoundError):
            container['test']



import sys

import os
import pytest

PY3 = sys.version_info[0] >= 3

if os.environ.get('MOCK_ATTRS'):
    @pytest.fixture(autouse=True, scope="function")
    def hide_attr(monkeypatch):
        """
        Hide attrs from Antidote as it is currently a dependency of pytest.
        """
        # Store original __import__
        builtin_import = __import__

        def _import(name, *args):
            if name == 'attr':
                raise ImportError(name)
            return builtin_import(name, *args)

        if PY3:
            builtins_module = 'builtins'
        else:
            builtins_module = '__builtin__'

        monkeypatch.setattr('{}.__import__'.format(builtins_module), _import)



import pytest

from antidote import DependencyManager


def test_attrs():
    manager = DependencyManager()
    container = manager.container

    try:
        import attr
    except ImportError:
        with pytest.raises(RuntimeError):
            manager.attrib()
        return

    @manager.register
    class Service(object):
        pass

    container['parameter'] = object()

    @attr.s
    class Test(object):
        service = manager.attrib(Service)
        parameter = manager.attrib(use_name=True)

    test = Test()

    assert container[Service] is test.service
    assert container['parameter'] is test.parameter

    @attr.s
    class BrokenTest(object):
        service = manager.attrib()

    with pytest.raises(ValueError):
        BrokenTest()



import pytest

from antidote.container import DependencyCycleError, InstantiationStack


class Service(object):
    pass


def test_format_stack():
    ds = InstantiationStack([Service, 'test', 1, Service])
    service_info = "{}.{}".format(Service.__module__, Service.__name__)

    assert service_info in ds.format_stack()
    assert "'test'" in ds.format_stack()
    assert " 1 " in ds.format_stack()

    assert ds.format_stack() in repr(ds)


def test_instantiating():
    ds = InstantiationStack()

    assert [] == list(ds)

    with ds.instantiating(InstantiationStack):
        with ds.instantiating('test'):
            assert [InstantiationStack, 'test'] == list(ds)

    assert [] == list(ds)

    with pytest.raises(DependencyCycleError):
        with ds.instantiating(InstantiationStack):
            with ds.instantiating(InstantiationStack):
                pass

    assert [] == list(ds)

    class CustomException(Exception):
        pass

    try:
        with ds.instantiating(InstantiationStack):
            raise CustomException()
    except CustomException:
        pass

    assert [] == list(ds)



from typing import Callable, Dict

from ..container import Dependency, Instance
from ..exceptions import (
    DependencyDuplicateError, DependencyNotProvidableError
)


class FactoryProvider:
    """
    Provider managing factories. When a dependency is requested, it tries
    to find a matching factory and builds it. Subclasses may also be built.
    """

    def __init__(self, auto_wire: bool = True) -> None:
        self.auto_wire = auto_wire
        self._factories = dict()  # type: Dict
        self._subclass_factories = dict()  # type: Dict

    def __repr__(self):
        return (
            "{}(auto_wire={!r}, factories={!r}, subclass_factories={!r}"
        ).format(
            type(self).__name__,
            self.auto_wire,
            tuple(self._factories.keys()),
            tuple(self._subclass_factories.keys())
        )

    def __antidote_provide__(self, dependency: Dependency) -> Instance:
        """
        Builds the dependency if a factory associated with the dependency_id
        can be found.

        Args:
            dependency: dependency to provide.

        Returns:
            A :py:class:`~.container.Instance` wrapping the built instance for
            the dependency.
        """
        try:
            factory = self._factories[dependency.id]
        except KeyError:
            for cls in getattr(dependency.id, '__mro__', []):
                try:
                    factory = self._subclass_factories[cls]
                    break
                except KeyError:
                    pass
            else:
                raise DependencyNotProvidableError(dependency.id)

        args = dependency.args
        if factory.takes_dependency_id:
            args = (dependency.id,) + args

        return Instance(factory(*args, **dependency.kwargs),
                        singleton=factory.singleton)

    def register(self,
                 dependency_id,
                 factory: Callable,
                 singleton: bool = True,
                 build_subclasses: bool = False):
        """
        Register a factory for a dependency.

        Args:
            dependency_id: ID of the dependency.
            factory: Callable used to instantiate the dependency.
            singleton: Whether the dependency should be mark as singleton or
                not for the :py:class:`~..container.DependencyContainer`.
            build_subclasses: If True, subclasses will also be build with this
                factory. If multiple factories are defined, the first in the
                MRO is used.
        """
        if not callable(factory):
            raise ValueError("The `factory` must be callable.")

        if not dependency_id:
            raise ValueError("`dependency_id` parameter must be specified.")

        dependency_factory = DependencyFactory(
            factory=factory,
            singleton=singleton,
            takes_dependency_id=build_subclasses,
        )

        if dependency_id in self._factories:
            raise DependencyDuplicateError(dependency_id)

        if build_subclasses:
            self._subclass_factories[dependency_id] = dependency_factory

        self._factories[dependency_id] = dependency_factory


class DependencyFactory(object):
    """
    Only used by the FactoryProvider, not part of the public API.

    Simple container to store information on how the factory has to be used.
    """
    __slots__ = ('factory', 'singleton', 'takes_dependency_id')

    def __init__(self,
                 factory,
                 singleton: bool,
                 takes_dependency_id: bool
                 ) -> None:
        self.factory = factory
        self.singleton = singleton
        self.takes_dependency_id = takes_dependency_id

    def __repr__(self):
        return (
            "{}(factory={!r}, singleton={!r}, takes_dependency_id={!r})"
        ).format(
            type(self).__name__,
            self.factory,
            self.singleton,
            self.takes_dependency_id
        )

    def __call__(self, *args, **kwargs):
        return self.factory(*args, **kwargs)



import pytest

from antidote import DependencyNotProvidableError
from antidote.providers.parameters import ParameterProvider, Dependency
from functools import reduce
from operator import getitem


def test_register():
    provider = ParameterProvider()
    conf = {
        'service': object(),
        'another': 1,
        'simple': 'yes',
        'recursive': {
            'test': object()
        },
        1: 1
    }

    def conf_parser(params, dependency_id):
        if isinstance(dependency_id, str):
            if dependency_id.startswith('conf:'):
                try:
                    return reduce(getitem, dependency_id[5:].split('.'),
                                  params)
                except TypeError as e:
                    raise LookupError(dependency_id) from e

        raise LookupError(dependency_id)

    provider.register(conf, conf_parser)

    def provide(e):
        return provider.__antidote_provide__(Dependency(e)).item

    with pytest.raises(DependencyNotProvidableError):
        provide('service')

    with pytest.raises(DependencyNotProvidableError):
        provide(1)

    assert conf['service'] is provide('conf:service')
    assert 1 == provide('conf:another')
    assert conf['recursive']['test'] is provide('conf:recursive.test')

    with pytest.raises(DependencyNotProvidableError):
        provide('conf:nothing')

    with pytest.raises(DependencyNotProvidableError):
        provide('conf:simple.yes.nothing')

    with pytest.raises(DependencyNotProvidableError):
        provide('conf:recursive.test.nothing')

    assert repr(conf) in repr(provider)



import functools
import typing
from itertools import islice
from typing import (
    Any, Callable, Dict, Iterable, Mapping, Sequence, Tuple,
    Union
)

from ._utils import get_arguments_specification
from .container import DependencyContainer, DependencyNotFoundError

_EMPTY_DEPENDENCY = object()

InjectionBlueprintType = Tuple[Tuple[str, bool, Any], ...]


class DependencyInjector:
    """
    Provides different methods to inject the dependencies from a container of
    any function.
    """

    def __init__(self, container: DependencyContainer) -> None:
        """Initialize the DependencyInjector.

        Args:
            container: :py:class:`~..DependencyContainer` from which to
                retrieve the dependencies.

        """
        self._container = container

    def __repr__(self):
        return "{}(container={!r}".format(
            type(self).__name__,
            self._container
        )

    def inject(self,
               func: Callable = None,
               arg_map: Union[Mapping, Iterable] = None,
               use_names: Union[bool, Iterable[str]] = False
               ) -> Callable:
        """
        Inject the dependency into the function lazily: they are only
        retrieved upon execution.

        Args:
            func: Callable for which the argument should be injected.
            arg_map: Custom mapping of the arguments name to their respective
                dependency id. A sequence of dependencies can also be
                specified, which will be mapped to the arguments through their
                order. Annotations are overriden.
            use_names: Whether the arguments name should be used to find for
                a dependency. An iterable of names may also be provided to
                restrict this to a subset of the arguments. Annotations are
                overriden, but not the arg_map.

        Returns:
            The decorator to be applied or the injected function if the
            argument :code:`func` was supplied.

        """

        def _inject(f):
            generate_args_kwargs = self._generate_args_kwargs
            injection_blueprint = self._generate_injection_blueprint(
                func=f,
                arg_map=arg_map or dict(),
                use_names=use_names
            )

            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                args, kwargs = generate_args_kwargs(
                    injection_blueprint=injection_blueprint,
                    args=args,
                    kwargs=kwargs
                )

                return f(*args, **kwargs)

            return wrapper

        return func and _inject(func) or _inject

    def bind(self,
             func: Callable = None,
             arg_map: Union[Mapping, Iterable] = None,
             use_names: Union[bool, Iterable[str]] = False,
             args: Sequence = None,
             kwargs: Dict = None
             ) -> Callable:
        """
        Creates a partial function with the injected arguments. It may be used
        whenever a function is called repeatedly, to diminish the runtime
        impact.

        Beware that this makes testing your function troublesome as you won't
        be able to change the dependencies used once imported.

        Args:
            func: Callable for which the argument should be injected.
            arg_map: Custom mapping of the arguments name to their respective
                dependency id. A sequence of dependencies can also be
                specified, which will be mapped to the arguments through their
                order. Annotations are overriden.
            use_names: Whether the arguments name should be used to find for
                a dependency. An iterable of names may also be provided to
                restrict this to a subset of the arguments. Annotations are
                overriden, but not the arg_map.
            args: Positional arguments passed on the function, overriding any
                injection, which will also be bound.
            kwargs: Keyword arguments passed on the function, overriding any
                injection, which will also be bound.

        Returns:
            The decorator to be applied or the injected function if the
            argument :code:`func` was supplied.

        """

        def _bind(f):
            new_args, new_kwargs = self._inject_into_arg_kwargs(
                func=f,
                arg_map=arg_map,
                use_names=use_names,
                args=args or tuple(),
                kwargs=kwargs or dict()
            )

            return functools.partial(f, *new_args, **new_kwargs)

        return func and _bind(func) or _bind

    def call(self,
             func: Callable,
             arg_map: Union[Mapping, Iterable] = None,
             use_names: Union[bool, Iterable[str]] = False,
             args: Sequence = None,
             kwargs: Dict = None
             ) -> Callable:
        """
        Call a function with specified arguments and keyword arguments.
        Dependencies are injected whenever necessary.

        Args:
            func: Callable for which the argument should be injected.
            arg_map: Custom mapping of the arguments name to their respective
                dependency id. A sequence of dependencies can also be
                specified, which will be mapped to the arguments through their
                order. Annotations are overriden.
            use_names: Whether the arguments name should be used to find for
                a dependency. An iterable of names may also be provided to
                restrict this to a subset of the arguments. Annotations are
                overriden, but not the arg_map.
            args: Positional arguments passed on the function, overriding any
                injection.
            kwargs: Keyword arguments passed on the function, overriding any
                injection.

        Returns:
            The decorator to be applied or the injected function if the
            argument :code:`func` was supplied.

        """

        new_args, new_kwargs = self._inject_into_arg_kwargs(
            func=func,
            arg_map=arg_map,
            use_names=use_names,
            args=args,
            kwargs=kwargs
        )

        return func(*new_args, **new_kwargs)

    def _inject_into_arg_kwargs(self,
                                func: Callable,
                                arg_map: Union[Mapping, Iterable] = None,
                                use_names: Union[bool, Iterable[str]] = False,
                                args: Sequence = None,
                                kwargs: Dict = None
                                ) -> Tuple[Sequence, Dict]:
        """
        Utility function to generate the injection blueprint and the new
        arguments in one step.
        """
        return self._generate_args_kwargs(
            injection_blueprint=self._generate_injection_blueprint(
                func=func,
                arg_map=arg_map,
                use_names=use_names
            ),
            args=args or tuple(),
            kwargs=kwargs or dict()
        )

    def _generate_args_kwargs(self,
                              injection_blueprint: InjectionBlueprintType,
                              args: Sequence,
                              kwargs: Dict
                              ) -> Tuple[Sequence, Dict]:
        """
        Generate the new arguments to be used by retrieving the missing
        dependencies based on the injection blueprint.

        If one argument has no default, is not set and is not mapped to a
        known dependency, :py:exc:`~..exceptions.DependencyNotFoundError` is
        raised.
        """
        kwargs = kwargs.copy()
        container = self._container

        for arg_name, has_default, dependency_id \
                in islice(injection_blueprint, len(args), None):
            if dependency_id is not _EMPTY_DEPENDENCY \
                    and arg_name not in kwargs:
                try:
                    kwargs[arg_name] = container[dependency_id]
                except DependencyNotFoundError:
                    if has_default is False:
                        raise

        return args, kwargs

    @staticmethod
    def _generate_injection_blueprint(func: Callable,
                                      arg_map: Union[Mapping, Iterable],
                                      use_names: Union[bool, Iterable[str]]
                                      ) -> InjectionBlueprintType:
        """
        Construct a list with all the necessary information about the arguments
        for dependency injection, named the injection blueprint. Storing it
        avoids significant execution overhead.

        The blueprint looks like:

        >>> [
        ...    (
        ...        'arg1',  # Name of the argument
        ...         False,  # whether it has a default value
        ...         'dependency id'  # associated dependency id
        ...     ),
        ...    ('arg2', True, _EMPTY_DEPENDENCY)
        ... ]

        """
        from collections import Mapping, Iterable

        try:
            annotations = typing.get_type_hints(func) or dict()
        except Exception:
            # Python 3.5.3 does not handle properly method wrappers
            annotations = dict()

        arg_spec, _, _ = get_arguments_specification(func)

        if arg_map is None:
            arg_to_dependency = {}  # type: Mapping
        elif isinstance(arg_map, Mapping):
            arg_to_dependency = arg_map
        elif isinstance(arg_map, Iterable):
            arg_to_dependency = {
                name: dependency_id
                for (name, _), dependency_id in zip(arg_spec, arg_map)
            }
        else:
            raise ValueError('Only a mapping or a iterable is supported for '
                             'arg_map, not {!r}'.format(arg_map))

        if use_names is None or use_names is False:
            use_names = set()
        elif use_names is True:
            use_names = set(e[0] for e in arg_spec)
        elif isinstance(use_names, Iterable):
            use_names = set(use_names)
        else:
            raise ValueError('Only an iterable or a boolean is supported for '
                             'use_names, not {!r}'.format(use_names))

        return tuple(
            (
                name,
                has_default,
                arg_to_dependency.get(
                    name,
                    name
                    if name in use_names else
                    annotations.get(name, _EMPTY_DEPENDENCY)
                ),
            )
            for name, has_default in arg_spec
        )



from configparser import RawConfigParser
from operator import getitem

import pytest

from antidote import (
    DependencyContainer, DependencyInjector, DependencyInstantiationError,
    DependencyManager, DependencyNotFoundError, DependencyNotProvidableError,
    Instance
)
from antidote.providers import FactoryProvider, ParameterProvider


def test_inject_bind():
    manager = DependencyManager()
    container = manager.container

    class Service(object):
        pass

    manager.register(Service)

    @manager.inject(bind=True)
    def f(x):
        return x

    with pytest.raises(TypeError):
        f()

    @manager.inject(arg_map=dict(x=Service), bind=True)
    def g(x):
        return x

    assert container[Service] is g()

    # arguments are bound, so one should not be able to pass injected
    # argument.
    with pytest.raises(TypeError):
        g(1)

    container['service'] = container[Service]
    container['service_bis'] = container[Service]

    @manager.inject(use_names=True, bind=True)
    def h(service, service_bis=None):
        return service, service_bis

    result = h()
    assert container[Service] is result[0]
    assert container[Service] is result[1]

    @manager.inject(use_names=('service',), bind=True)
    def h(service, service_bis=None):
        return service, service_bis

    result = h()
    assert container[Service] is result[0]
    assert None is result[1]


def test_inject_with_mapping():
    manager = DependencyManager()
    container = manager.container

    class Service(object):
        pass

    manager.register(Service)

    class AnotherService(object):
        pass

    manager.register(AnotherService)

    @manager.inject
    def f(x):
        return x

    with pytest.raises(TypeError):
        f()

    @manager.inject(arg_map=dict(x=Service))
    def g(x):
        return x

    assert container[Service] is g()

    @manager.inject(arg_map=(Service,))
    def g(x):
        return x

    assert container[Service] is g()

    @manager.inject(arg_map=dict(x=Service))
    def h(x, b=1):
        return x

    assert container[Service] is h()

    @manager.inject(arg_map=(Service,))
    def h(x, b=1):
        return x

    assert container[Service] is h()

    manager.arg_map = dict(x=Service, y=AnotherService)

    @manager.inject
    def u(x, y):
        return x, y

    assert container[Service] is u()[0]
    assert container[AnotherService] is u()[1]

    @manager.inject(arg_map=dict(y=Service))
    def v(x, y):
        return x, y

    assert container[Service] is v()[0]
    assert container[Service] is v()[1]

    @manager.inject(arg_map=(AnotherService,))
    def v(x, y):
        return x, y

    assert container[AnotherService] is v()[0]
    assert container[AnotherService] is v()[1]


def test_wire():
    manager = DependencyManager()
    container = manager.container

    class Service(object):
        pass

    class AnotherService(object):
        pass

    manager.register(Service)
    manager.register(AnotherService)

    @manager.wire(arg_map=dict(service=Service,
                               another_service=AnotherService))
    class Something(object):
        def f(self, service):
            return service

        def g(self, another_service):
            return another_service

        def h(self, service, another_service):
            return service, another_service

        def u(self):
            pass

        def v(self, nothing):
            return nothing

    something = Something()
    assert container[Service] is something.f()
    assert container[AnotherService] is something.g()

    s1, s2 = something.h()
    assert container[Service] is s1
    assert container[AnotherService] is s2

    something.u()

    with pytest.raises(TypeError):
        something.v()


def test_use_names():
    manager = DependencyManager(use_names=False)
    container = manager.container

    _service = object()
    _service_bis = object()
    container['service'] = _service
    container['service_bis'] = _service_bis

    def f(service):
        return service

    with pytest.raises(TypeError):
        manager.inject(f)()

    def g(service, service_bis=None, something=None):
        return service, service_bis, something

    g_result = manager.inject(use_names=True)(g)()
    assert _service is g_result[0]
    assert _service_bis is g_result[1]
    assert None is g_result[2]

    g_result = manager.inject(use_names=('service',))(g)()
    assert _service is g_result[0]
    assert None is g_result[1]
    assert None is g_result[2]

    # use names for every injection by default.
    manager.use_names = True

    assert _service is manager.inject(f)()

    with pytest.raises(TypeError):
        manager.inject(use_names=False)(f)()


def test_register():
    manager = DependencyManager(auto_wire=True)
    container = manager.container

    @manager.register
    class Service(object):
        pass

    assert isinstance(container[Service], Service)

    @manager.register(arg_map=dict(service=Service))
    class AnotherService(object):
        def __init__(self, service):
            self.service = service

    assert isinstance(container[AnotherService], AnotherService)
    assert container[Service] is container[AnotherService].service
    # singleton
    assert container[AnotherService] is container[AnotherService]

    container['service'] = object()

    @manager.register(use_names=True)
    class YetAnotherService(object):
        def __init__(self, service):
            self.service = service

    assert isinstance(container[YetAnotherService], YetAnotherService)
    assert container['service'] is container[YetAnotherService].service
    # singleton
    assert container[YetAnotherService] is container[YetAnotherService]

    @manager.register(singleton=False)
    class SingleUsageService(object):
        pass

    assert isinstance(container[SingleUsageService], SingleUsageService)
    assert container[SingleUsageService] is not container[SingleUsageService]

    @manager.register(auto_wire=False)
    class BrokenService(object):
        def __init__(self, service):
            self.service = service

    with pytest.raises(DependencyInstantiationError):
        container[BrokenService]

    @manager.register(auto_wire=('__init__', 'method'),
                      arg_map=dict(service=Service,
                                   x=SingleUsageService,
                                   yet=YetAnotherService))
    class ComplexWiringService(object):
        def __init__(self, service):
            self.service = service

        def method(self, x, yet):
            return x, yet

    assert isinstance(container[ComplexWiringService], ComplexWiringService)
    assert container[Service] is container[ComplexWiringService].service
    output = container[ComplexWiringService].method()

    assert isinstance(output[0], SingleUsageService)
    assert output[1] is container[YetAnotherService]


def test_register_non_class():
    manager = DependencyManager()

    with pytest.raises(ValueError):
        manager.register(object())

    def f():
        pass

    with pytest.raises(ValueError):
        manager.register(f)


def test_factory_function():
    manager = DependencyManager()
    container = manager.container

    class Service(object):
        pass

    with pytest.raises(ValueError):
        @manager.factory
        def faulty_service_provider():
            return Service()

    @manager.factory(dependency_id=Service)
    def service_provider():
        return Service()

    assert isinstance(container[Service], Service)
    # is a singleton
    assert container[Service] is container[Service]

    class AnotherService(object):
        def __init__(self, service):
            self.service = service

    @manager.factory(arg_map=dict(service=Service),
                     dependency_id=AnotherService)
    def another_service_provider(service):
        return AnotherService(service)

    assert isinstance(container[AnotherService], AnotherService)
    # is a singleton
    assert container[AnotherService] is container[AnotherService]
    assert isinstance(container[AnotherService].service, Service)

    s = object()
    container['test'] = s

    class YetAnotherService:
        pass

    @manager.factory(use_names=True, dependency_id=YetAnotherService)
    def injected_by_name_provider(test):
        return test

    assert s is container[YetAnotherService]

    with pytest.raises(ValueError):
        manager.factory(1)

    with pytest.raises(ValueError):
        @manager.factory
        class Test:
            pass


def test_factory_class():
    manager = DependencyManager()
    container = manager.container

    class Service(object):
        pass

    with pytest.raises(ValueError):
        @manager.factory
        class FaultyServiceProvider(object):
            def __call__(self):
                return Service()

    @manager.factory(dependency_id=Service)
    class ServiceProvider(object):
        def __call__(self):
            return Service()

    assert isinstance(container[Service], Service)
    # is a singleton
    assert container[Service] is container[Service]

    class AnotherService(object):
        def __init__(self, service):
            self.service = service

    @manager.factory(arg_map=dict(service=Service),
                     dependency_id=AnotherService)
    class AnotherServiceProvider(object):
        def __init__(self, service):
            self.service = service
            assert isinstance(service, Service)

        def __call__(self, service):
            assert self.service is service
            return AnotherService(service)

    assert isinstance(container[AnotherService], AnotherService)
    # is a singleton
    assert container[AnotherService] is container[AnotherService]
    assert isinstance(container[AnotherService].service, Service)

    container['test'] = object()

    class YetAnotherService(object):
        pass

    @manager.factory(use_names=True, dependency_id=YetAnotherService)
    class YetAnotherServiceProvider(object):
        def __init__(self, test):
            self.test = test

        def __call__(self, test):
            assert self.test is test
            return test

    assert container['test'] is container[YetAnotherService]

    class OtherService(object):
        pass

    @manager.factory(use_names=True,
                     arg_map=dict(service=Service),
                     auto_wire=('__init__',),
                     dependency_id=OtherService)
    class OtherServiceProvider(object):
        def __init__(self, test, service):
            self.test = test
            self.service = service

        def __call__(self):
            return self.test, self.service

    output = container[OtherService]
    assert output[0] is container['test']
    assert isinstance(output[1], Service)


def test_provider():
    manager = DependencyManager()
    container = manager.container

    container['service'] = object()

    @manager.provider(use_names=True)
    class DummyProvider(object):
        def __init__(self, service=None):
            self.service = service

        def __antidote_provide__(self, dependency):
            if dependency.id == 'test':
                return Instance(dependency.id)
            else:
                raise DependencyNotProvidableError(dependency)

    assert isinstance(container.providers[DummyProvider], DummyProvider)
    assert container.providers[DummyProvider].service is container['service']
    assert 'test' == container['test']

    with pytest.raises(DependencyNotFoundError):
        container['test2']

    with pytest.raises(ValueError):
        manager.provider(object())

    with pytest.raises(ValueError):
        @manager.provider
        class MissingAntidoteProvideMethod(object):
            pass

    with pytest.raises(TypeError):
        @manager.provider(auto_wire=False)
        class MissingDependencyProvider(object):
            def __init__(self, service):
                self.service = service

            def __antidote_provide__(self, dependency):
                return Instance(dependency.id)


def test_providers():
    manager = DependencyManager()

    assert 2 == len(manager.providers)
    assert FactoryProvider in manager.providers
    assert ParameterProvider in manager.providers

    @manager.provider
    class DummyProvider(object):
        def __antidote_provide__(self, dependency):
            return Instance(1)

    assert DummyProvider in manager.providers


def test_parameters():
    manager = DependencyManager()
    container = manager.container

    manager.register_parameters({'test1': 'some value'}, getter=getitem)
    assert 'some value' == container['test1']

    manager.register_parameters({'test2': 'another value'}, getter=getitem,
                                prefix='conf:')
    assert 'another value' == container['conf:test2']

    manager.register_parameters({'test3': {'nested': 'yes'}}, getter=getitem,
                                split='.')
    assert 'yes' == container['test3.nested']

    manager.register_parameters({'param': '1', 'paramb': {'test': 2}},
                                getter=getitem, prefix='params:', split='.')

    assert '1' == container['params:param']
    assert 2 == container['params:paramb.test']

    with pytest.raises(DependencyNotFoundError):
        container[object()]

    with pytest.raises(DependencyNotFoundError):
        container['test3.nested.yes']


def test_register_parameters_custom_getter():
    manager = DependencyManager()
    container = manager.container

    @manager.register_parameters({'a': {'b': {'c': 99}}})
    def parser(obj, item):
        from functools import reduce
        if isinstance(item, str):
            return reduce(getitem, list(item), obj)

        raise LookupError(item)

    assert 99 == container['abc']

    # Does not fail with missing dependency
    with pytest.raises(DependencyNotFoundError):
        container[object()]


def test_invalid_arguments():
    manager = DependencyManager()

    with pytest.raises(ValueError):
        manager.register_parameters(object(), getter=object())

    with pytest.raises(ValueError):
        manager.register_parameters(object(), prefix=object())

    with pytest.raises(ValueError):
        manager.register_parameters(object(), split=object())


def test_parameters_with_configparser():
    manager = DependencyManager()
    container = manager.container

    cfg = RawConfigParser()
    cfg.add_section('test')
    cfg.set('test', 'param', '100')

    manager.register_parameters(cfg, getter=getitem, split='.')

    assert '100' == container['test.param']

    with pytest.raises(DependencyNotFoundError):
        container['section.option']

    with pytest.raises(DependencyNotFoundError):
        container['test.option']

    with pytest.raises(DependencyNotFoundError):
        container['test.param.test']


def test_manager_repr():
    manager = DependencyManager()

    assert repr(manager.container) in repr(manager)
    assert repr(manager.injector) in repr(manager)
    assert 'auto_wire' in repr(manager)
    assert 'use_names' in repr(manager)
    assert 'mapping' in repr(manager)


def test_context():
    manager = DependencyManager()
    manager.container['param'] = 1

    with manager.context(include=[]):
        with pytest.raises(DependencyNotFoundError):
            manager.container['param']

        manager.container[DependencyInjector]
        manager.container[DependencyContainer]

    with manager.context(include=['param']):
        assert 1 == manager.container['param']

        manager.container[DependencyInjector]
        manager.container[DependencyContainer]

    with manager.context(exclude=['param']):
        with pytest.raises(DependencyNotFoundError):
            manager.container['param']

        manager.container[DependencyInjector]
        manager.container[DependencyContainer]

    with manager.context(missing=['param']):
        with pytest.raises(DependencyNotFoundError):
            manager.container['param']

        manager.container[DependencyInjector]
        manager.container[DependencyContainer]

    with manager.context(dependencies={'param': 2}):
        assert 2 == manager.container['param']

        manager.container[DependencyInjector]
        manager.container[DependencyContainer]



import pytest

from antidote import (
    DependencyContainer, DependencyInjector,
    DependencyNotFoundError
)


@pytest.fixture()
def injector():
    container = DependencyContainer()
    injector = DependencyInjector(container)
    return injector


def test_bind(injector: DependencyInjector):
    def f(a, b):
        return a, b

    # arguments properly passed on
    a, b = injector.bind(f, args=(1, 2))()
    assert 1 == a and 2 == b

    a, b = injector.bind(f, args=(1,), kwargs=dict(b=2))()
    assert 1 == a and 2 == b

    a, b = injector.bind(f, kwargs=dict(a=1, b=2))()
    assert 1 == a and 2 == b


def test_call(injector: DependencyInjector):
    def f(a, b):
        return a, b

    # arguments properly passed on
    a, b = injector.call(f, args=(1, 2))
    assert 1 == a and 2 == b

    a, b = injector.call(f, args=(1,), kwargs=dict(b=2))
    assert 1 == a and 2 == b

    a, b = injector.call(f, kwargs=dict(a=1, b=2))
    assert 1 == a and 2 == b


def test_arg_map(injector: DependencyInjector):
    container = injector._container

    class Service(object):
        pass

    container[Service] = Service()

    def f(service):
        return service

    inject_mapping = injector.inject(arg_map=dict(service=Service))
    inject_sequence = injector.inject(arg_map=(Service,))

    # function called properly
    assert container[Service] is inject_mapping(f)()
    assert container[Service] is inject_sequence(f)()

    # function called properly
    assert inject_mapping(f)(None) is None
    assert inject_sequence(f)(None) is None

    def g(service, parameter=2):
        return service, parameter

    # argument still passed on
    assert (container[Service], 2) == inject_mapping(g)()
    assert (container[Service], 2) == inject_sequence(g)()

    class Obj(object):
        def __init__(self, service, parameter=2):
            self.service = service
            self.parameter = parameter

    class UnknownService(object):
        pass

    # faulty mapping
    inject_mapping = injector.inject(arg_map=dict(service=UnknownService))
    inject_sequence = injector.inject(arg_map=dict(service=UnknownService))
    with pytest.raises(DependencyNotFoundError):
        inject_mapping(f)()

    with pytest.raises(DependencyNotFoundError):
        inject_sequence(f)()

    # with no mapping, raises the same as with no arguments
    inject_mapping = injector.inject()
    with pytest.raises(TypeError):
        inject_mapping(f)()


def test_use_names(injector: DependencyInjector):
    container = injector._container

    container['test'] = object()

    def f(test):
        return test

    # test is inject by name
    f = injector.inject(func=f, use_names=True)
    assert container['test'] == f()

    container['yes'] = 'yes'
    container['no'] = 'no'

    def g(yes, no=None):
        return yes, no

    inject = injector.inject(use_names=('yes',))
    assert (container['yes'], None) == inject(g)()


def test_defaults(injector: DependencyInjector):
    container = injector._container

    container['service'] = object()

    def f(service, optional_service=None):
        return service, optional_service

    # test is inject by name
    inject = injector.inject(use_names=True)

    assert (container['service'], None) == inject(f)()

    container['optional_service'] = object()
    assert (container['service'], container['optional_service']) == inject(f)()


def test_method_wrapper_type_hints_error(monkeypatch,
                                         injector: DependencyInjector):
    def raises(*args, **kwargs):
        raise TypeError()

    monkeypatch.setattr('typing.get_type_hints', raises)

    def f():
        pass

    injector.inject(f)()


def test_repr(injector):
    assert repr(injector._container) in repr(injector)


@pytest.mark.parametrize(
    'method_name',
    [
        'bind',
        'inject',
        'call'
    ]
)
def test_value_error(method_name, injector: DependencyInjector):
    def f():
        pass

    with pytest.raises(ValueError):
        getattr(injector, method_name)(func=f, use_names=object())

    with pytest.raises(ValueError):
        getattr(injector, method_name)(func=f, arg_map=object())


def test_injection_order(injector: DependencyInjector):
    container = injector._container

    class Service:
        pass

    A = object()
    B = Service()
    C = object()
    container['A'] = A
    container[Service] = B
    container['C'] = C

    def f(A: Service):
        return A

    assert B is injector.inject(f)()
    assert A is injector.inject(f, use_names=True)()
    assert C is injector.inject(f, arg_map=['C'])()
    assert C is injector.inject(f, use_names=True, arg_map=['C'])()



import pytest

from antidote import (
    DependencyDuplicateError, DependencyNotProvidableError
)
from antidote.providers.factories import (
    DependencyFactory, FactoryProvider, Dependency
)


class Service(object):
    def __init__(self, *args):
        pass


class ServiceSubclass(Service):
    pass


class AnotherService(object):
    def __init__(self, *args):
        pass


@pytest.fixture()
def provider():
    return FactoryProvider()


def test_dependency_factory():
    o = object()

    def test(*args, **kwargs):
        return o, args, kwargs

    df = DependencyFactory(test, True, False)

    assert repr(test) in repr(df)
    assert (o, (1,), {'param': 'none'}) == df(1, param='none')


def test_register(provider: FactoryProvider):
    provider.register(Service, Service)

    dependency = provider.__antidote_provide__(Dependency(Service))
    assert isinstance(dependency.item, Service)
    assert repr(Service) in repr(provider)


def test_register_factory_id(provider: FactoryProvider):
    provider.register(Service, lambda: Service())

    dependency = provider.__antidote_provide__(Dependency(Service))
    assert isinstance(dependency.item, Service)


def test_singleton(provider: FactoryProvider):
    provider.register(Service, Service, singleton=True)
    provider.register(AnotherService, AnotherService, singleton=False)

    provide = provider.__antidote_provide__
    assert provide(Dependency(Service)).singleton is True
    assert provide(Dependency(AnotherService)).singleton is False


def test_register_for_subclasses(provider: FactoryProvider):
    provider.register(Service, lambda cls: cls(), build_subclasses=True)

    assert isinstance(
        provider.__antidote_provide__(Dependency(Service)).item,
        Service
    )
    assert isinstance(
        provider.__antidote_provide__(Dependency(ServiceSubclass)).item,
        Service
    )

    with pytest.raises(DependencyNotProvidableError):
        provider.__antidote_provide__(Dependency(AnotherService))


def test_register_not_callable_error(provider: FactoryProvider):
    with pytest.raises(ValueError):
        provider.register(1, 1)


def test_register_id_null(provider: FactoryProvider):
    with pytest.raises(ValueError):
        provider.register(None, Service)


def test_duplicate_error(provider: FactoryProvider):
    provider.register(Service, Service)

    with pytest.raises(DependencyDuplicateError):
        provider.register(Service, Service)

    with pytest.raises(DependencyDuplicateError):
        provider.register(Service, lambda: Service())



from antidote import DependencyManager


def test_inject():
    manager = DependencyManager()
    container = manager.container

    class Service(object):
        pass

    manager.register(Service)

    @manager.inject
    def f(service: Service):
        return service

    assert container[Service] is f()


def test_register():
    manager = DependencyManager()
    container = manager.container

    @manager.register
    class Service(object):
        pass

    assert isinstance(container[Service], Service)

    @manager.register
    class AnotherService(object):
        def __init__(self, service: Service):
            self.service = service

    assert isinstance(container[AnotherService], AnotherService)
    assert isinstance(container[AnotherService].service, Service)


def test_provider_function():
    manager = DependencyManager()
    container = manager.container

    @manager.register
    class Service(object):
        pass

    class AnotherService(object):
        def __init__(self, service):
            self.service = service

    @manager.factory
    def provider(service: Service) -> AnotherService:
        return AnotherService(service)

    assert isinstance(container[AnotherService], AnotherService)
    assert isinstance(container[AnotherService].service, Service)


def test_provider_class():
    manager = DependencyManager()
    container = manager.container

    @manager.register
    class Service(object):
        pass

    class AnotherService(object):
        def __init__(self, service):
            self.service = service

    @manager.factory
    class Provider(object):
        def __call__(self, service: Service) -> AnotherService:
            return AnotherService(service)

    assert isinstance(container[AnotherService], AnotherService)
    assert isinstance(container[AnotherService].service, Service)



import pytest

from antidote._utils import get_arguments_specification


def f(a, b, c=1):
    pass


def g(a, *args):
    pass


def h(b=None, **kwargs):
    pass


def k():
    pass


class Dummy(object):
    def f(self, a, b=1, *args, **kwargs):
        pass

    @classmethod
    def g(cls, a):
        pass

    @staticmethod
    def h():
        pass


d = Dummy()


@pytest.mark.parametrize(
    'func,expected',
    [
        (f, ([('a', False), ('b', False), ('c', True)], False, False)),
        (g, ([('a', False)], True, False)),
        (h, ([('b', True)], False, True)),
        (k, ([], False, False)),
        (Dummy.f, ([('self', False), ('a', False), ('b', True)], True, True)),
        (Dummy.g, ([('a', False)], False, False)),
        (Dummy.h, ([], False, False)),
        (d.f, ([('a', False), ('b', True)], True, True)),
        (d.g, ([('a', False)], False, False)),
        (d.h, ([], False, False)),
    ],
    ids=[
        'f', 'g', 'h', 'k',
        'cls.f', 'cls.g', 'cls.h',
        'instance.f', 'instance.g', 'instance.h'
    ]
)
def test_arg_spec(func, expected):
    assert expected == get_arguments_specification(func)



import contextlib
import threading
from collections import OrderedDict
from typing import Any, Union, Mapping, Iterable

from .stack import InstantiationStack
from ..exceptions import (
    DependencyCycleError, DependencyInstantiationError,
    DependencyNotFoundError, DependencyNotProvidableError
)

_SENTINEL = object()


class DependencyContainer:
    """
    Container of dependencies which are instantiated lazily by providers.
    Singleton are cached to ensure they're not rebuilt more than once.

    One can specify additional arguments on how to build a dependency, by
    requiring a :py:class:`~Dependency` or using :py:meth:`~provide`.

    Neither :code:`__contains__()` nor :code:`__delitem__()` are implemented as
    they are error-prone, they would only operate on the cache, not the set of
    available dependencies.
    """

    def __init__(self):
        self.providers = OrderedDict()
        self._cache = {}
        self._cache_lock = threading.RLock()
        self._instantiation_stack = InstantiationStack()

    def __str__(self):
        return "{}(providers=({}))".format(
            type(self).__name__,
            ", ".join("{}={}".format(name, p)
                      for name, p in self.providers.items()),
        )

    def __repr__(self):
        return "{}(providers=({}), cache={!r})".format(
            type(self).__name__,
            ", ".join("{!r}={!r}".format(name, p)
                      for name, p in self.providers.items()),
            self._cache
        )

    def __getitem__(self, dependency):
        """
        Get the specified dependency. :code:`item` is either the dependency_id
        or a :py:class:`~Dependency` instance in order to provide additional
        arguments to the providers.
        """
        try:
            return self._cache[dependency]
        except KeyError:
            pass

        try:
            with self._cache_lock, \
                    self._instantiation_stack.instantiating(dependency):
                try:
                    return self._cache[dependency]
                except KeyError:
                    pass

                for provider in self.providers.values():
                    try:
                        instance = provider.__antidote_provide__(
                            dependency
                            if isinstance(dependency, Dependency) else
                            Dependency(dependency)
                        )  # type: Instance
                    except DependencyNotProvidableError:
                        pass
                    else:
                        if instance.singleton:
                            self._cache[dependency] = instance.item

                        return instance.item

        except DependencyCycleError:
            raise

        except Exception as e:
            raise DependencyInstantiationError(dependency) from e

        raise DependencyNotFoundError(dependency)

    def provide(self, *args, **kwargs):
        """
        Utility method which creates a :py:class:`~Dependency` and passes it to
        :py:meth:`~__getitem__`.
        """
        return self[Dependency(*args, **kwargs)]

    def __setitem__(self, dependency_id, dependency):
        """
        Set a dependency in the cache.
        """
        with self._cache_lock:
            self._cache[dependency_id] = dependency

    def update(self, *args, **kwargs):
        """
        Update the cached dependencies.
        """
        with self._cache_lock:
            self._cache.update(*args, **kwargs)

    @contextlib.contextmanager
    def context(self,
                dependencies: Union[Mapping, Iterable] = None,
                include: Iterable = None,
                exclude: Iterable = None,
                missing: Iterable = None
                ):
        """
        Creates a context within one can control which of the defined
        dependencies available or not. Any changes will be discarded at the
        end.

        Args:
            dependencies: Dependencies instances used to override existing ones
                in the new context.
            include: Iterable of dependency to include. If None
                everything is accessible.
            exclude: Iterable of dependency to exclude.
            missing: Iterable of dependency which should raise a
                :py:exc:`~.exceptions.DependencyNotFoundError` even if a
                provider could instantiate them.

        """

        with self._cache_lock:
            original_cache = self._cache

            if missing:
                missing = set(missing)
                exclude = set(exclude) | missing if exclude else missing

                class Cache(dict):
                    def __missing__(self, key):
                        if key in missing:
                            raise DependencyNotFoundError(key)

                        raise KeyError(key)

                self._cache = Cache(original_cache if not include else {})

            elif include is None:
                self._cache = original_cache.copy()
            else:
                self._cache = {}

            if include:
                for dependency in include:
                    self._cache[dependency] = original_cache[dependency]

            if exclude:
                for dependency in exclude:
                    try:
                        del self._cache[dependency]
                    except KeyError:
                        pass

            if dependencies:
                self._cache.update(dependencies)

        try:
            yield
        finally:
            with self._cache_lock:
                self._cache = original_cache


class Dependency:
    """
    Simple container which can be used to specify a dependency ID with
    additional arguments, :code:`args` and :code:`kwargs`, for the provider.

    If no additional arguments are provided it is equivalent to the unwrapped
    dependency id.

    >>> from antidote import antidote, Dependency
    >>> antidote.container['name'] = 'Antidote'
    >>> antidote.container[Dependency('name')]
    'Antidote'

    """
    __slots__ = ('id', 'args', 'kwargs')

    def __init__(self, *args, **kwargs):
        self.id = args[0]
        assert not isinstance(self.id, Dependency)
        self.args = args[1:]
        self.kwargs = kwargs

    def __repr__(self):
        return "{}({!r}, *{!r}, **{!r})".format(
            type(self).__name__,
            self.id,
            self.args,
            self.kwargs
        )

    def __hash__(self):
        if self.args or self.kwargs:
            try:
                # Try most precise hash first
                return hash((
                    self.id,
                    self.args,
                    tuple(self.kwargs.items())
                ))
            except TypeError:
                # If type error, return the best error-free hash possible
                return hash((
                    self.id,
                    len(self.args),
                    tuple(self.kwargs.keys())
                ))

        return hash(self.id)

    def __eq__(self, other):
        return (
            (
                not self.kwargs and not self.args
                and (self.id is other or self.id == other)
            )
            or (
                isinstance(other, Dependency)
                and (self.id is other.id or self.id == other.id)
                and self.args == other.args
                and self.kwargs == other.kwargs
            )
        )


class Instance:
    """
    Simple wrapper which has to be used by providers when returning an
    instance of a dependency.

    This enables the container to know if the returned dependency needs to
    be cached or not (singleton).
    """

    __slots__ = ('item', 'singleton')

    def __init__(self, item: Any, singleton: bool = False) -> None:
        self.item = item
        self.singleton = singleton

    def __repr__(self):
        return "{}(instance={!r}, singleton={!r})".format(
            type(self).__name__,
            self.item,
            self.singleton
        )



import inspect
from collections import deque
from contextlib import contextmanager
from typing import Iterable

from ..exceptions import DependencyCycleError


class InstantiationStack:
    """
    Stores the stack of dependency instantiation to detect and prevent cycles
    by raising DependencyCycleError.

    Used in the DependencyContainer.

    This class is not thread-safe by itself.
    """

    def __init__(self, stack: Iterable = None) -> None:
        self._stack = deque(stack or [])
        self._dependencies = set(self._stack)

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            self.format_stack()
        )

    def __iter__(self):
        return iter(self._stack)

    def format_stack(self, sep: str = ' => ') -> str:
        """
        Returns a human readable representation of the current stack.
        """
        return sep.join(map(self._format_dependency_id, self._stack))

    @classmethod
    def _format_dependency_id(cls, dependency_id) -> str:
        if inspect.isclass(dependency_id):
            return "{}.{}".format(dependency_id.__module__,
                                  dependency_id.__name__)

        return repr(dependency_id)

    @contextmanager
    def instantiating(self, dependency_id):
        """
        Context Manager which has to be used when instantiating the
        dependency to keep track of the dependency path.

        When a cycle is detected, a DependencyCycleError is raised.
        """
        if dependency_id in self._dependencies:
            self._stack.append(dependency_id)
            message = self.format_stack()
            self._clear()
            raise DependencyCycleError(message)

        self._stack.append(dependency_id)
        self._dependencies.add(dependency_id)
        try:
            yield
        except Exception:
            self._clear()
            raise
        else:
            self._dependencies.remove(self._stack.pop())

    def _clear(self):
        self._stack = deque()
        self._dependencies = set()



import threading
import time

import random

from antidote import DependencyManager


class Service(object):
    pass


class AnotherService(object):
    pass


def make_delayed_factory(a, b, factory):
    def f():
        time.sleep(a + b * random.random())
        return factory()

    return f


def test_instantiation_safety():
    n_threads = 10
    manager = DependencyManager()

    manager.factory(
        make_delayed_factory(.1, .1, Service),
        dependency_id=Service,
        singleton=True
    )
    manager.factory(
        make_delayed_factory(.1, .1, AnotherService),
        dependency_id=AnotherService,
        singleton=False
    )

    singleton_got = []
    non_singleton_got = []

    def worker():
        singleton_got.append(manager.container[Service])
        non_singleton_got.append(manager.container[AnotherService])

    threads = [
        threading.Thread(target=worker)
        for _ in range(n_threads)
    ]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()

    assert 1 == len(set(singleton_got))
    assert n_threads == len(set(non_singleton_got))



import os
import sys
import shutil
from codecs import open

from setuptools import setup, find_packages


here = os.path.dirname(os.path.abspath(__file__))

about = {}
with open(os.path.join(here, 'src', 'antidote', '__version__.py')) as f:
    exec(f.read(), about)

# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    print("Removing previous builds...")
    if os.path.exists(os.path.join(here, 'dist')):
        shutil.rmtree(os.path.join(here, 'dist'))

    print("Building distribution...")
    os.system('python setup.py sdist bdist_wheel')

    print("Uploading the package to PyPi with Twine...")
    os.system('twine upload dist/*')

    print("Pushing git tags")
    os.system('git tag v{0}'.format(about['__version__']))
    os.system('git push --tags')

    print("Done !")
    sys.exit()


with open('README.rst', 'r') as f:
    readme = f.read()

setup(
    name='antidote',
    version=about['__version__'],
    description='Transparent dependency injection.',
    long_description=readme,
    author='Benjamin Rabier',
    url='https://github.com/Finistere/antidote',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=[],
    extras_require={
        ":python_version<'3.5'": ["typing"],
        "docs": [
            "sphinx",
            "sphinx_autodoc_typehints",
            "sphinx-rtd-theme",
            "attrs"
        ],
        "tests": [
            "pytest",
            "pytest-cov"
        ]
    },
    license='MIT',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ),
    keywords='dependency injection',
    zip_safe=False,
)



import pytest

from antidote import Dependency


class Service:
    pass


tests = [
    ('test', (1,), {}),
    (1, tuple(), {'test': 1}),
    (Service, (1, 'test'), {'another': 'no'}),
    (Service, tuple(), {}),
    (Service, tuple(), {'not_hashable': {'hey': 'hey'}})
]


@pytest.mark.parametrize('id,args,kwargs', tests)
def test_eq_hash(id, args, kwargs):
    p = Dependency(id, *args, **kwargs)

    # does not fail
    hash(p)

    for f in (lambda e: e, hash):
        assert (f(Dependency(id, **kwargs)) == f(p)) is not len(args)
        assert (f(Dependency(id, *args)) == f(p)) is not len(kwargs)
        assert (f(Dependency(id)) == f(p)) is not (len(args)
                                                   or len(kwargs))
        assert (f(id) == f(p)) is not (len(args)
                                       or len(kwargs))

    assert repr(id) in repr(p)
    assert repr(args) in repr(p)
    assert repr(kwargs) in repr(p)



import inspect
from typing import Callable, Sequence, Tuple

ArgSpecType = Tuple[Sequence[Tuple[str, bool]], bool, bool]


def get_arguments_specification(func: Callable) -> ArgSpecType:
    """
    Extract the name and if a default is set for each argument.
    """
    arguments = []
    has_var_positional = False
    has_var_keyword = False

    for name, parameter in inspect.signature(func).parameters.items():
        if parameter.kind is parameter.VAR_POSITIONAL:
            has_var_positional = True
            continue

        if parameter.kind is parameter.VAR_KEYWORD:
            has_var_keyword = True
            continue

        arguments.append((name,
                          parameter.default is not parameter.empty))

    return arguments, has_var_positional, has_var_keyword



from antidote import DependencyManager


def test_attrs():
    manager = DependencyManager()
    container = manager.container

    try:
        import attr
    except ImportError:
        return

    @manager.register
    class Service(object):
        pass

    @attr.s
    class Test(object):
        service: Service = manager.attrib()

    test = Test()

    assert container[Service] is test.service
